/*
 * Copyright (C) 2014 University of Stuttgart, IKR
 *
 * This file is part of IKR tcp-mux-client as available from
 * http://www.ikr.uni-stuttgart.de/IKRSimLib.
 *
 * This file is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (GPL)
 * as published by the Free Software Foundation, in version 2 as it
 * comes in the file "COPYING" in directory doc of IKR tcp-mux-client.
 *
 * IKR tcp-mux-client is distributed in the hope that it will be useful, but
 * without any warranty of any kind. Please contact the IKR if you
 * need additional information (http://www.ikr.uni-stuttgart.de).
 *
 * Alternatively you may license IKR tcp-mux-client under a different
 * (non-free) license. Please contact the IKR for further details.
 *
 */

#include "tcp-mux-client.h"
#include <err.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <string.h>
#include <sys/select.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/sockios.h>
#include <termios.h>
#include <linux/serial.h>
#include <alloca.h>
#include <sys/mman.h>
#include <netinet/in.h>
#include <sys/epoll.h>
#include <stdint.h>


#ifdef DO_UTP
#include <utp.h>
#endif

#ifdef DO_CRC
#include "crc.h"
#define DIGEST_LENGTH 4
#else
#define DIGEST_LENGTH 0
#endif

#define MAX_ALLOCA_SIZE (4*4096)

struct header {
	uint32_t type;
	uint32_t length;
} header;

#define AUTO_READ_DEFAULT true
#define AUTO_ACCEPT_DEFAULT true
#define UTP_BUF_SIZE 1024*1024
#define UTP_VERSION 2

typedef struct listentry listentry;

typedef void (*readHandler)(listentry *);
typedef void (*writeHandler)(listentry *);

enum Protocol {tcp, utp, control, udp};

#ifdef DO_UTP
struct utp_listentry {
	utp_socket * sock;  		/* utp_socket* */
	utp_context* utp_ctx; 		/* just used for utp-related sockets: the udp "master" socket and the utp_sockets. */
	bool autoRead; 				/* whether to read data automatically */
	bool autoAccept;			/* whether to accept new connections automatically */
	struct utp_listentry * next;	/* organized as linked list*/
};
typedef struct utp_listentry utp_listentry;
utp_listentry* utp_head = NULL ;

struct utp_ctx_listentry {
	utp_context* utp_ctx;
	struct utp_ctx_listentry * next;	/* organized as linked list*/
};
typedef struct utp_ctx_listentry utp_ctx_listentry;
utp_ctx_listentry * utp_ctx_head = NULL;


static utp_listentry * create_new_utp_listentry(utp_socket* sock, utp_context * ctx) {
	utp_listentry * uentry;
	uentry = malloc(sizeof(utp_listentry));
	uentry->sock = sock;
	uentry->utp_ctx = ctx;
	uentry->autoRead = AUTO_READ_DEFAULT;
	uentry->autoAccept = AUTO_ACCEPT_DEFAULT;
	uentry->next = utp_head;
	utp_head = uentry;
	return uentry;
}

static utp_listentry * uentry_from_usock(utp_socket * usock) {
	utp_listentry * e;
	for (e = utp_head; e != NULL; e = e->next) {
		if (e->sock == usock)
			return e;
	}
	return 0;
}

static void delete_utplistentry(utp_listentry * uentry) {
	utp_listentry * prev = utp_head;
	if (prev == uentry)
		utp_head = utp_head->next;
	else {
		while (prev->next != uentry && prev->next != NULL)
			prev = prev->next;
		if (prev->next == NULL) {
			fprintf(stderr, "list entry not found in %s\n", __func__);
			fflush(stderr);
			exit(1);
		} else {
			prev->next = prev->next->next;
		}
	}
	free(uentry);
}

#endif

struct listentry {
	int fd;  					/* file descriptor of the given entry if tcp, utp_sock* if utp */
	int prot; 					/* protocol identifier to adopt handling in handlers */
	readHandler readHandler;	/* function used to handle a possible read */
	writeHandler writeHandler;	/* function used to handle a possible write (socket ready) */
	bool autoRead; 				/* whether to read data automatically */
	bool autoAccept;			/* whether to accept new connections automatically */
#ifdef DO_UTP
	utp_context* utp_ctx; 		/* just used for utp-related sockets: the udp "master" socket and the utp_sockets. */
#endif
	int epoll_flags;			/* guess what... */
	struct listentry * next;	/* organized as linked list*/
};

listentry* tcp_head;
int controlfd;
int epollfd;

static listentry * create_new_listentry(int fd, int prot) {
	listentry * entry;
	entry = malloc(sizeof(listentry));
	entry->fd = fd;
	entry->prot = prot;
	entry->next = tcp_head;
	tcp_head = entry;

	entry->autoRead = AUTO_READ_DEFAULT;
	entry->autoAccept = AUTO_ACCEPT_DEFAULT;
	entry->readHandler = NULL;
	entry->writeHandler = NULL;

	entry->epoll_flags = 0;
	struct epoll_event ev;
	ev.events = 0;
	ev.data.ptr = entry;
	if (epoll_ctl(epollfd, EPOLL_CTL_ADD, fd, &ev) == -1) {
		err(1, "epoll_ctl() failed");
	}

	return entry;
}

static void delete_listentry(listentry * e) {
	/*
	 * closed fd will be removed from the epoll set automatically
	 */
	listentry * prev = tcp_head;
	if (prev == e)
		tcp_head = tcp_head->next;
	else {
		while (prev->next != e && prev->next != NULL)
			prev = prev->next;
		if (prev->next == NULL) {
			fprintf(stderr, "list entry not found in %s\n", __func__);
			fflush(stderr);
			exit(1);
		} else {
			prev->next = prev->next->next;
		}
	}
	free(e);
}

static void set_handlers(listentry * e, readHandler rh, writeHandler wh) {
	e->readHandler = rh;
	e->writeHandler = wh;
	int new_flags = 0;
	if (rh != NULL)
		new_flags |= (EPOLLIN | EPOLLRDHUP);
	if (wh != NULL)
		new_flags |= EPOLLOUT;
	if (new_flags != e->epoll_flags) {
		struct epoll_event ev;
		ev.events = new_flags;
		ev.data.ptr = e;
		if (epoll_ctl(epollfd, EPOLL_CTL_MOD, e->fd, &ev) == -1) {
			err(1, "epoll_ctl() failed (old %x new %x)", e->epoll_flags, new_flags);
		}
		e->epoll_flags = new_flags;
	}
}

static inline void set_write_handler(listentry * e, writeHandler wh) {
	set_handlers(e, e->readHandler, wh);
}

static inline void set_read_handler(listentry * e, readHandler rh) {
	set_handlers(e, rh, e->writeHandler);
}

static listentry * entry_from_fd(uint32_t fd) {
	listentry * e;
	for (e = tcp_head; e != NULL; e = e->next) {
		if (e->fd == fd)
			return e;
	}
	fprintf(stderr, "invalid file descriptor %d, list entry not found\n", fd);
	fflush(stderr);
	exit(-1);
}

#ifdef DO_UTP
static int udp_socket_from_utp_context(utp_context * uctx){
	listentry * e;
	for (e = tcp_head; e != NULL ; e = e->next) {
			if (e->utp_ctx == uctx)
				if(e->prot == udp){
					return e->fd;
				}
	}
	fprintf(stderr, "could not find udp socket for utp_context %d,\n", (int)(intptr_t)uctx);
	unsigned int i = 0;
	if(tcp_head == NULL)
		fprintf(stderr, "entry list is empty!!");
	for (e = tcp_head; e != NULL; e = e->next) {
		fprintf(stderr, "Entry %i: prot = %i, utp_ctx = %d", i, e->prot, (int)(intptr_t)e->utp_ctx);
	}
	fflush(stderr);
	exit(-1);
}

//static listentry * entry_from_utp_socket(utp_socket * usock){
//	listentry * e;
//	for (e = head; e->next != NULL; e = e->next) {
//		if (e->utp_sock == usock)
//			return e;
//	}
//	fprintf(stderr, "invalid UTP socket %d, list entry not found\n", usock);
//	fflush(stderr);
//	exit(-1);
//}
#endif

#ifdef DO_CRC
static crc_t calcCRC(const unsigned char * data1, size_t size1, const unsigned char * data2, size_t size2) {
	crc_t crc;
	crc = crc_init();
	crc = crc_update(crc, data1, size1);
	crc = crc_update(crc, data2, size2);
	crc = crc_finalize(crc);
	return crc;
}

static bool checkCRC(const unsigned char * data1, size_t size1, const unsigned char * data2, size_t size2) {
	crc_t crc = calcCRC(data1, size1, data2, size2);

	crc_t crc_in_msg = *((crc_t *) (data2 + size2));
	if (crc_in_msg != crc) {
		printf("message digest mismatch: digest is %d but should be %d\n", crc_in_msg, crc);
		fflush(stdout);
		return false;
	}
	//printf("CRC OK\n");
	//fflush(stdout);
	return true;
}
#endif

static void assert_length(int length, int target, const char* method) {
	if (length != target) {
		fprintf(stderr, "message length mismatch (%s, should be %d, is %d)\n", method, target, length);
		fflush(stderr);
		exit(-1);
	}
}

/*
 * Insert the default parts of the message
 */
static void insert_type_fd(int32_t msg_type, uint32_t fd, char * msg) {
	(*(int32_t *) (msg + 0)) = msg_type;
	(*(uint32_t *) (msg + 8)) = fd;
}

/*
 * Insert the length into the message.
 * Only count the used part (without basic parts)
 */
static void insert_length(int32_t length, char * msg) {
	(*(int32_t *) (msg + 4)) = (int32_t) (length + FD_SIZE);
}

// writes data to the controlling device
static void control_write(const char* data, int length) {
	int written = 0;
	int retval;

	while (written < length) {
		retval = write(controlfd, data + written, length - written);
		if (retval >= 0) {
			written += retval;
		} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
			// ignore
		} else {
			err(1, "control_write() failed");
		}
	}

#ifdef DO_CRC
	crc_t crc = calcCRC((unsigned char *) data, sizeof(header), (unsigned char *)(data + sizeof(header)), length - sizeof(header));
	unsigned char * crc_dat = (unsigned char *) &crc;

	written = 0;
	while (written < DIGEST_LENGTH) {
		retval = write(controlfd, crc_dat + written, DIGEST_LENGTH - written);
		if (retval >= 0) {
			written += retval;
		} else if (errno == EAGAIN || errno == EWOULDBLOCK) {
			// ignore
		} else {
			err(1, "control_write() failed");
		}
	}
#endif
}

#ifdef DO_UTP
static uint64 utp_callback_sendto(utp_callback_arguments *a);
static uint64 utp_callback_on_error(utp_callback_arguments *a);
static uint64 utp_callback_on_state_change(utp_callback_arguments *a);
static uint64 utp_callback_on_read(utp_callback_arguments *a);
static uint64 utp_callback_on_firewall(utp_callback_arguments *a);
static uint64 utp_callback_on_accept(utp_callback_arguments *a);

static utp_context* setup_utp_context(const uint32_t bufsize, const uint32_t targetUs) {
	utp_context* ctx = utp_init(UTP_VERSION, bufsize, targetUs);
	utp_set_callback(ctx, UTP_LOG,				NULL);
	utp_set_callback(ctx, UTP_SENDTO,			&utp_callback_sendto);
	utp_set_callback(ctx, UTP_ON_ERROR,			&utp_callback_on_error);
	utp_set_callback(ctx, UTP_ON_STATE_CHANGE,	&utp_callback_on_state_change);
	utp_set_callback(ctx, UTP_ON_READ,			&utp_callback_on_read);
	utp_set_callback(ctx, UTP_ON_FIREWALL,		&utp_callback_on_firewall);
	utp_set_callback(ctx, UTP_ON_ACCEPT,		&utp_callback_on_accept);
	utp_ctx_listentry* ctxentry;
	ctxentry = malloc(sizeof(utp_ctx_listentry));
	ctxentry->utp_ctx = ctx;
	ctxentry->next = utp_ctx_head;
	utp_ctx_head = ctxentry;
	return ctx;
}

static uint64 utp_callback_sendto(utp_callback_arguments *a){
	// need to find udp socket for _this_ utp_socket (utp pseudo socket)
	int fd = udp_socket_from_utp_context(a->context);
	sendto(fd, a->buf, a->len, 0, a->address, a->address_len);
	return 0;
}
#endif

/*
 * Create a default message without any additional value
 */
static void send_trigger_msg(int32_t msg_type, uint32_t fd) {
	char msg[TRIGGER_MSG_SIZE];
	insert_type_fd(msg_type, fd, msg);
	insert_length(0, msg);
	control_write(msg, TRIGGER_MSG_SIZE);
}

static void mux_getname_internal(int fd, bool peer, char buf[], unsigned int * length);

/*
 * Called to notify the simlib of the real fd which is associated to a temporary fd.
 * After this message has been sent, the tmpFd will not be used any more.
 */
static void notify_real_fd(uint32_t tmpFd, uint32_t fd) {
	/* allocate for ipv6, should be sufficient for v4, too */
	unsigned int sockname_len = 3 * sizeof(int32_t) + sizeof(struct in6_addr);
	char msg[TRIGGER_MSG_SIZE + sizeof(int32_t) + sockname_len];

	(*(uint32_t *) (msg + TRIGGER_MSG_SIZE)) = fd;
	mux_getname_internal(fd, false, msg + TRIGGER_MSG_SIZE + 4, &sockname_len);

	insert_type_fd(MUX_MSG_CREATED_SOCK, tmpFd, msg);
	insert_length(4 + sockname_len, msg);

	control_write(msg, TRIGGER_MSG_SIZE + 4 + sockname_len);
}

static void handle_socket_read(listentry * e);

static inline int min(int a, int b) {
	if(a < b)
		return a;
	else
		return b;
}

/*
 * called when the data should be read from the socket
 */
static int mux_read_internal(listentry * e, int32_t max_amount_to_read) {
	size_t retval;
	char * msg;
	int do_free = false;

	int offset = TRIGGER_MSG_SIZE + (2 * sizeof(uint32_t));
	int initial_read_size = MAX_ALLOCA_SIZE - offset;
	if (max_amount_to_read != -1)
		initial_read_size = min(initial_read_size, max_amount_to_read);
	int msg_size = initial_read_size + offset;
	msg = alloca(msg_size);

	insert_type_fd(MUX_MSG_READ, e->fd, msg);

	int32_t length = 0;
	errno = -1;
	if (e->prot == tcp) {
		retval = read(e->fd, msg + offset, initial_read_size);
		if (retval >= 0) {
			length = retval;
		} else {
			if (errno != EAGAIN) {
				err(1, "first read of %d bytes from TCP socket failed (%s)", initial_read_size, strerror(errno));
			}
		}
		// read more data
		if (retval >= 0 && length == initial_read_size && (max_amount_to_read > initial_read_size || max_amount_to_read == -1)) {
			// determine how much is left
			int remaining_bytes = 0;
			if (ioctl(e->fd, SIOCINQ, &remaining_bytes) != 0)
				err(1, "ioctl() failed\n");
			if (max_amount_to_read != -1)
				remaining_bytes = min(remaining_bytes, max_amount_to_read - initial_read_size);
			if (remaining_bytes > 0) {
				// copy data to larger buffer
				char * smallmsg = msg;
				msg = malloc(msg_size + remaining_bytes);
				if (msg == NULL) {
					fprintf(stderr, "malloc() failed\n");
					fflush(stderr);
					exit(1);
				}
				do_free = true;
				memcpy(msg, smallmsg, msg_size);
				// read remaining bytes
				retval = read(e->fd, msg + msg_size, remaining_bytes);
				if (retval >= 0) {
					length += retval;
					retval = length; // to be signalled to simlib, see below
				} else {
					// retval stays negative, signalled below
					if (errno != EAGAIN) {
						err(1, "second read failed (%s)", strerror(errno));
					}
				}
			}
		}
		// send reply message
		insert_length(length + 2 * sizeof(uint32_t), msg);
		*((size_t*) (msg + TRIGGER_MSG_SIZE)) = retval;
		*((int32_t*) (msg + TRIGGER_MSG_SIZE + sizeof(uint32_t))) = errno;
		control_write(msg, TRIGGER_MSG_SIZE + length + 2 * sizeof(uint32_t));

		if (do_free)
			free(msg);
#ifdef DO_UTP
	} else if (e->prot == udp) {
		// hoping that we never can read more than initial_read_size from a UTP socket because it would be discarded contrary to the TCP socket
		struct sockaddr_in src_addr;
		socklen_t addrlen = sizeof(src_addr);
		byte * udpmsg = alloca(initial_read_size);
		retval = recvfrom(e->fd, udpmsg, initial_read_size, MSG_DONTWAIT, (struct sockaddr *)&src_addr, &addrlen);
		if (retval < 0) {
			if (errno == EAGAIN || errno == EWOULDBLOCK) {
				utp_issue_deferred_acks(e->utp_ctx);
			} else {
				err(1, "reading %d bytes from UDP-socket for UTP failed (%s)", initial_read_size, strerror(errno));
			}
		}
		length = utp_process_udp(e->utp_ctx, udpmsg, retval, (struct sockaddr *)&src_addr, addrlen);
		if (read < 0) {
			err(1, "Message read from UDP-socket for UTP has not been recognized as UTP.");
		}
		utp_issue_deferred_acks(e->utp_ctx);
		utp_check_timeouts(e->utp_ctx);
		// be ready to read again
		set_read_handler(e, handle_socket_read);
		if(length == 0)
			length++; // make sure, this read return that it read something
		//TODO:calc length
#endif
	} else {
		err(1, "mux_read_internal called for non-tcp & non-utp-socket (prot is %d)", e->prot);
	}
	return length;
}

#ifdef DO_UTP
static uint64 utp_callback_on_read(utp_callback_arguments *a) {

	// send reply message
	char * msg;
	size_t offset = TRIGGER_MSG_SIZE + (2 * sizeof(uint32_t));
	size_t msg_size = a->len + offset;
	msg = alloca(msg_size);
	insert_type_fd(MUX_MSG_READ, (uint32_t) a->socket, msg);
	insert_length(a->len + 2 * sizeof(uint32_t), msg);
	//no retval and errno available
	*((size_t*) (msg + TRIGGER_MSG_SIZE)) = a->len;
	*((int32_t*) (msg + TRIGGER_MSG_SIZE + sizeof(uint32_t))) = 0;
	memcpy(msg + offset, a->buf, a->len);
	control_write(msg, msg_size);
	utp_read_drained(a->socket);
//	utp_issue_deferred_acks(a->context);
	return 0;
}

static uint64 utp_callback_on_firewall(utp_callback_arguments *a) {
	return 0;
}

static uint64 utp_callback_on_accept(utp_callback_arguments *a) {
	utp_listentry * uentry;
	utp_socket* usock = a->socket;
	// need to create a new entry for the utp-pseudo-socket utp_socket
	//make sure there is no collision
	for (uentry = utp_head; uentry != NULL && uentry->next != NULL; uentry = uentry->next) {
		if (uentry->sock == usock){
			err(1, "when accepting UTP connection fd with utp_sock* already exists!");
		}
	}
	// dont really need a listentry because its all managed by callbacks but...
	uentry = create_new_utp_listentry(usock, a->context); // ia32 only

	/* allocate for ipv6, should be sufficient for v4, too */
	unsigned int sockname_len = 3 * sizeof(int32_t) + sizeof(struct in6_addr);
	char sockname_buf[sockname_len];
	mux_getname_internal((int) usock, false, sockname_buf, &sockname_len);

	/* allocate for ipv6, should be sufficient for v4, too */
	unsigned int peername_len = 3 * sizeof(int32_t) + sizeof(struct in6_addr);
	char peername_buf[peername_len];
	mux_getname_internal((int) usock, true, peername_buf, &peername_len);

	// send reply message
	int msglen = TRIGGER_MSG_SIZE + 2 * sizeof(uint32_t) + sockname_len + peername_len;
	char msg[msglen];
	insert_type_fd(MUX_MSG_ACCEPT, (uint32_t)udp_socket_from_utp_context(a->context), msg);
	insert_length(msglen - TRIGGER_MSG_SIZE, msg);
	*((size_t*) (msg + TRIGGER_MSG_SIZE)) = (intptr_t)usock;
	*((size_t*) (msg + TRIGGER_MSG_SIZE + sizeof(uint32_t))) = errno;

	memcpy(msg + TRIGGER_MSG_SIZE + 2 * sizeof(uint32_t), sockname_buf, sockname_len);
	memcpy(msg + TRIGGER_MSG_SIZE + 2 * sizeof(uint32_t) + sockname_len, peername_buf, peername_len);

	control_write(msg, msglen);

	// signal control that socket is writable
	send_trigger_msg(MUX_MSG_WRITABLE, (intptr_t)usock);
	return 0;
}

static uint64 utp_callback_on_error(utp_callback_arguments *a) {
	err(1, "Error: %s\n", utp_error_code_names[a->error_code]);
	utp_close(a->socket);
	return 0;
}

static uint64 utp_callback_on_state_change(utp_callback_arguments *a) {
	switch (a->state) {
		case UTP_STATE_CONNECT:
			;
			char msg[TRIGGER_MSG_SIZE + 4]; //additional int32_t for errorcode
			insert_type_fd(MUX_MSG_CONNECT, (intptr_t)(a->socket), msg);
			insert_length(4, msg);
			(*(int32_t *) (msg + TRIGGER_MSG_SIZE)) = 0;
			control_write(msg, TRIGGER_MSG_SIZE + 4);
			send_trigger_msg(MUX_MSG_WRITABLE, (intptr_t)(a->socket));
			break;
		case UTP_STATE_WRITABLE:
			send_trigger_msg(MUX_MSG_WRITABLE, (intptr_t)(a->socket));
			break;

		case UTP_STATE_EOF:
			utp_close(a->socket);
			// remove entry from list. ignore for now.
			break;
	}

	return 0;
}
#endif

static void mux_read(unsigned char *buf, int buflen) {
	assert_length(buflen, 8, "socket read");
	listentry * e = entry_from_fd(*((uint32_t *) buf));
	//reactivate the read consideration
	set_read_handler(e, handle_socket_read);

	//read the data
	int32_t max_amount_to_read = *((int32_t *) (buf + 4));

	mux_read_internal(e, max_amount_to_read);
}

/*
 * called when a sockets fd becomes ready for writing
 */
static void handle_socket_write(listentry * e) {
	// don't signal again until simlib wrote
	set_write_handler(e, NULL);

	//send message "socket ready for writing" to simlib
	send_trigger_msg(MUX_MSG_WRITABLE, e->fd);
}

/*
 * called when a sockets fd becomes ready for reading
 *
 * forwards the availability of readable data
 */
static void handle_socket_read(listentry * e) {
	if (e->autoRead) {
		int bytes = mux_read_internal(e, -1);
		if (bytes == 0) {
			// remote disconnect???
			set_read_handler(e, NULL);
		}
	} else {
		// remove read handler to avoid multiple calls while not handled
		set_read_handler(e, NULL);

		send_trigger_msg(MUX_MSG_DATA_AVAILABLE, e->fd);
	}
}

static void mux_accept_internal(listentry * e);

/*
 * called when a connection is tried to established with listening socket
 */
static void handle_new_listener(listentry * e) {
	if (e->autoAccept) {
		mux_accept_internal(e);
	} else {
		//remove read handler
		set_read_handler(e, NULL);
		send_trigger_msg(MUX_MSG_CLIENT_AVAILABLE, e->fd);
	}
}

/*
 * called when a connecting socket becomes ready for writing (-> is now connected)
 */
static void handle_socket_connected(listentry * e) {
	// connection has not been established yet
	int32_t errorcode;
	socklen_t size = sizeof(errorcode);
	int retval = getsockopt(e->fd, SOL_SOCKET, SO_ERROR, &errorcode, &size);
	if (retval != 0)
		err(1, "getsockopt() failed");

	// if established then set connecting to false
	if (errorcode == 0) {
		// write handler is now ready
		// read handler can now be used (if set immediately after connect(), read() would return EOF in case of an error
		set_handlers(e, handle_socket_read, handle_socket_write);
	}

	char msg[TRIGGER_MSG_SIZE + 4]; //additional int32_t for errorcode
	insert_type_fd(MUX_MSG_CONNECT, e->fd, msg);
	insert_length(4, msg);
	(*(int32_t *) (msg + TRIGGER_MSG_SIZE)) = errorcode;
	control_write(msg, TRIGGER_MSG_SIZE + 4);
}

/*
 * opens a new TCP connection and creates the corresponding
 * entries in the list. should send the fd back to the
 * controlling dev as reference
 *
 * buf is msg data as received from control dev
 */
static void mux_tcp_socket_connect(unsigned char * buf, int length) {
	struct sockaddr addr;
	memset(&addr, 0, sizeof(struct sockaddr));
	int sock;
	int retval;
	listentry * entry;
	uint32_t tmpFd = *((uint32_t *) buf);

	// connect to an IPv4 address
	if (length == 12) {
		// 4 (tmpFd) + 4 (ip) + 4 (port) => 12 bytes
		assert_length(length, 12, "open_tcp_connectionv4()");

		// open socket (SOCK_NONBLOCK is not supported by old 2.6.26 kernel, therefore we do that manually)
		sock = socket(AF_INET, SOCK_STREAM, 0);
		if (sock == -1)
			err(1, "socket() error, create client socket (IPv4)");

        retval = fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK);
        if (retval != 0)
           err(1, "fcntl() error, set NONBLOCK flag");

		/* All IP addresses are exchanged in Big Endian == Java VM Byte Order  == Network Byte Order. No conversion required.
		 * Ports have to be exchanged in host byte order, conversion needed.
		 * */
		struct sockaddr_in* addrv4 = (struct sockaddr_in*) &addr;
		addrv4->sin_family = AF_INET;
		addrv4->sin_addr = *((struct in_addr *) (buf + sizeof(uint32_t))); //buf + 4
		addrv4->sin_port = ntohs(*((uint16_t *) (buf + sizeof(uint32_t) + sizeof(struct in_addr)))); //buf + 4+4
		retval = connect(sock, (const struct sockaddr *) &addr, sizeof(struct sockaddr_in));
	} else { //should only be IPv6 up to now :-)
		// connect to an IPv6 address
		// 4(tmpFd) + 4 (port) + 16 (local addr)
		assert_length(length, sizeof(uint32_t) + sizeof(uint32_t) + sizeof(struct in6_addr), "open_tcp_connectionv6()");

		// open socket (SOCK_NONBLOCK is not supported by old 2.6.26 kernel, therefore we do that manually)
		sock = socket(AF_INET6, SOCK_STREAM, 0);
		if (sock == -1)
			err(1, "socket() error, create client socket (IPv6)");

        retval = fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK);
        if (retval != 0)
           err(1, "fcntl() error, set NONBLOCK flag");

		/* All IP addresses are exchanged in Big Endian == Java VM Byte Order  == Network Byte Order. No conversion required.
		 * Ports have to be exchanged in host byte order, conversion needed.
		 * */
		struct sockaddr_in6* addrv6 = (struct sockaddr_in6*) &addr;
		addrv6->sin6_family = AF_INET6;
		addrv6->sin6_addr = *((struct in6_addr*) (buf + sizeof(uint32_t))); //buf + 4
		addrv6->sin6_port = ntohs(*((uint16_t *) (buf + sizeof(uint32_t) + sizeof(struct in6_addr)))); //buf + 4+16
		retval = connect(sock, (const struct sockaddr *) &addr, sizeof(struct sockaddr_in6));
	}

	if (retval == -1) {
		// connecting in progress
		if (errno != EINPROGRESS) {
			printf("connect returns %i\n", errno);
			err(1, "connect() error, create client");
		}
	}

	// create new entry and append it to the list
	entry = create_new_listentry(sock, tcp);
	// read handler set in handle_socket_connected
	set_write_handler(entry, handle_socket_connected);

	// notify simlib of newly created socket
	notify_real_fd(tmpFd, sock);
}

#ifdef DO_UTP
/*
 * opens a new UTP connection and creates the corresponding
 * entries in the list. should send the fd back to the
 * controlling dev as reference
 *
 * buf is msg data as received from control dev:
 * // 4 (tmpFd) + 4 buffersize + 4 (ip) + 4 (port) => 16 bytes
 */
static void mux_utp_socket_connect(unsigned char * buf, int length) {
	struct sockaddr addr;
	memset(&addr, 0, sizeof(struct sockaddr));
	int sock;
	int retval;
	listentry * entry;
	uint32_t tmpFd = *((uint32_t *) buf);
	unsigned char * buflength = buf + sizeof(uint32_t);
	unsigned char * targetUs = buflength + sizeof(uint32_t);
	unsigned char * addr4 = targetUs + sizeof(uint32_t);
	unsigned char * addr6 = addr4;

	utp_context * ctx = setup_utp_context(*(int*) buflength, *(int*) targetUs);
	utp_socket*  usock;

	// connect to an IPv4 address
	if (length == 20) {
		// 4 (tmpFd) + 4 (ip) + 4 (port) => 16 bytes
		assert_length(length, 20, "mux_utp_socket_connect_IPv4");

		unsigned char * port4 = addr4 + sizeof(struct in_addr);

		// open socket (SOCK_NONBLOCK is not supported by old 2.6.26 kernel, therefore we do that manually)
		sock = socket(AF_INET,  SOCK_DGRAM, IPPROTO_UDP);
		if (sock < 0)
			err(1, "socket() error, create utp client socket (IPv4)");
		int on = 1;
		if (setsockopt(sock, SOL_IP, IP_RECVERR, &on, sizeof(on)) != 0)
			err(1, "setsockopt() error, set IP_RECVERR flag");

		usock = utp_create_socket(ctx);

		// create new entries for udp socket and utp-pseudo-socket and append them to the list
		// the udp-entry MUST be available before calling utp_connect!!!
		entry = create_new_listentry(sock, udp);
		entry->utp_ctx = ctx;
		set_read_handler(entry, handle_socket_read);
		create_new_utp_listentry(usock, ctx);

		/* All IP addresses are exchanged in Big Endian == Java VM Byte Order  == Network Byte Order. No conversion required.
		 * Ports have to be exchanged in host byte order, conversion needed.
		 * */
		struct sockaddr_in* addrv4 = (struct sockaddr_in*) &addr;
		addrv4->sin_family = AF_INET;
		addrv4->sin_addr = *((struct in_addr *) addr4); //buf + 8
		addrv4->sin_port = ntohs(*((uint16_t *) port4)); //buf + 4+4
		retval = utp_connect(usock, &addr, sizeof(struct sockaddr_in));
	} else { //should only be IPv6 up to now :-)
		unsigned char * port6 = addr6 + sizeof(struct in6_addr);

		// connect to an IPv6 address
		// 4(tmpFd) + 4 buffersize + 4 (port) + 16 (local addr)
		assert_length(length, sizeof(uint32_t) + sizeof(uint32_t) + sizeof(uint32_t) +sizeof(uint32_t) + sizeof(struct in6_addr), "mux_utp_socket_connect_IPv6");

		// open socket (SOCK_NONBLOCK is not supported by old 2.6.26 kernel, therefore we do that manually)
		sock = socket(AF_INET6, SOCK_STREAM, 0);
		if (sock == -1)
			err(1, "socket() error, create client socket (IPv6)");
		int on = 1;
		if (setsockopt(sock, SOL_IP, IP_RECVERR, &on, sizeof(on)) != 0)
			err(1, "setsockopt() error, set IP_RECVERR flag");

		usock = utp_create_socket(ctx);
		// create new entries for udp socket and utp-pseudo-socket and append them to the list
		// the udp-entry MUST be available before calling utp_connect!!!
		entry = create_new_listentry(sock, udp);
		entry->utp_ctx = ctx;
		set_read_handler(entry, handle_socket_read);
		create_new_utp_listentry(usock, ctx);

		/* All IP addresses are exchanged in Big Endian == Java VM Byte Order  == Network Byte Order. No conversion required.
		 * Ports have to be exchanged in host byte order, conversion needed.
		 * */
		struct sockaddr_in6* addrv6 = (struct sockaddr_in6*) &addr;
		addrv6->sin6_family = AF_INET6;
		addrv6->sin6_addr = *((struct in6_addr*) addr6); //buf + 4 + 4
		addrv6->sin6_port = ntohs(*((uint16_t *) port6)); //buf + 4 + 4 + 16
		retval = utp_connect(usock, &addr, sizeof(struct sockaddr_in6));
	}
	if (retval < 0) {
		// connecting in progress
		if (errno != EINPROGRESS) {
			printf("connect returns %i\n", errno);
			err(1, "connect() error, create client");
		}
	}

	// notify simlib of newly created socket
	notify_real_fd(tmpFd, (int)usock);
}
#endif

static void mux_tcp_socket_bind_listen(unsigned char * buf, int length) {
	int sock;
	int retval;
	int bind_errno;
	listentry * entry;
	uint32_t tmpFd = *((uint32_t *) buf);

	// bind to an IPv4 address
	if (length == 12) {
		// 4(tmpFd) + 4 (port) + 4 (local addr)
		assert_length(length, 12, "create_server_v4");
		/* All IP addresses are exchanged in Big Endian == Java VM Byte Order  == Network Byte Order. No conversion required.
		 * Ports have to be exchanged in host byte order, conversion needed.
		 * */
		struct sockaddr_in addrv4;
		memset(&addrv4, 0, sizeof(struct sockaddr_in));
		addrv4.sin_family = AF_INET;
		addrv4.sin_addr = *((struct in_addr *) (buf + sizeof(uint32_t))); //buf + 4
		addrv4.sin_port = ntohs(*((uint16_t *) (buf + sizeof(uint32_t) + sizeof(struct in_addr)))); //buf + 4+4
		// open socket (SOCK_NONBLOCK is not supported by old 2.6.26 kernel, therefore we do that manually)
		sock = socket(AF_INET, SOCK_STREAM, 0);
		if (sock == -1)
			err(1, "socket() error, create server");
        retval = fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK);
        if (retval != 0)
           err(1, "fcntl() error, set NONBLOCK flag");
		// bind address
		retval = bind(sock, (const struct sockaddr *)&addrv4, sizeof(struct sockaddr_in));
		bind_errno = errno;
	} else {
		// bind to an IPv6 address
		// 4(tmpFd) + 4 (port) + 16 (local addr)
		assert_length(length, 24, "create_server_v6");
		/* All IP addresses are exchanged in Big Endian == Java VM Byte Order  == Network Byte Order. No conversion required.
		 * Ports have to be exchanged in host byte order, conversion needed.
		 * */
		struct sockaddr_in6 addrv6;
		memset(&addrv6, 0, sizeof(struct sockaddr_in6));
		addrv6.sin6_family = AF_INET6;
		addrv6.sin6_addr = *((struct in6_addr*) (buf + sizeof(uint32_t))); //buf + 4
		addrv6.sin6_port = ntohs(*((uint16_t *) (buf + sizeof(uint32_t) + sizeof(struct in6_addr)))); //buf + 4+16
		// open socket (SOCK_NONBLOCK is not supported by old 2.6.26 kernel, therefore we do that manually)
		sock = socket(AF_INET6, SOCK_STREAM, 0);
		if (sock == -1)
			err(1, "socket() error, create TCP server socket");
        retval = fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK);
        if (retval != 0)
           err(1, "fcntl() error, set NONBLOCK flag");
		// bind address
		retval = bind(sock, (const struct sockaddr *)&addrv6, sizeof(struct sockaddr_in6));
		bind_errno = errno;
	}

	// notify simlib of newly created socket
	notify_real_fd(tmpFd, sock);

	char msg[TRIGGER_MSG_SIZE + 8];
	insert_type_fd(MUX_MSG_BIND_RET, sock, msg);
	insert_length(8, msg);
	(*(uint32_t *) (msg + TRIGGER_MSG_SIZE)) = retval;
	(*(uint32_t *) (msg + TRIGGER_MSG_SIZE + 4)) = bind_errno;
	control_write(msg, TRIGGER_MSG_SIZE + 8);

	if (retval == 0) {
		// listen to it
		retval = listen(sock, 1024); // max backlock == 1024, shouldn't make a difference as each connection is accepted immediately
		if (retval != 0) {
			err(1, "socket() error, listening failed");
		}

		// create new entry and append it to the list
		entry = create_new_listentry(sock, tcp);
		set_read_handler(entry, handle_new_listener);
	}
}

#ifdef DO_UTP
/* set up UTP listening socket.
 * Works by opening a "master" udp socket and initializing an utp_context.
 * With patched utplib, the buffersize can be used as parameter so buf looks like:
 * 4 Byte offset, 4 Byte (int) buffsize, 4 Byte IPv4 addr, 4 Byte Port)
 */
static void mux_utp_socket_bind_listen(unsigned char * buf, int length) {
	int sock;
	int retval;
	int bind_errno;
	listentry * entry;
	uint32_t tmpFd = *((uint32_t *) buf);
	unsigned char * buflength = buf + sizeof(uint32_t);
	unsigned char * targetUs = buflength + sizeof(uint32_t);
	unsigned char * addr = targetUs + sizeof(uint32_t);
	unsigned char * port = addr + sizeof(struct in_addr);

	utp_context * ctx = setup_utp_context( *(int*) buflength, *(int*) targetUs);
	// bind to an IPv4 address
	if (length == 20) {
		// 4(tmpFd) + 4 (port) + 4 (local addr)
		assert_length(length, 20, "mux_utp_socket_bind_listen_v4");
		/* All IP addresses are exchanged in Big Endian == Java VM Byte Order  == Network Byte Order. No conversion required.
		 * Ports have to be exchanged in host byte order, conversion needed.
		 * */
		struct sockaddr_in addrv4;
		memset(&addrv4, 0, sizeof(struct sockaddr_in));
		addrv4.sin_family = AF_INET;
		addrv4.sin_addr = *((struct in_addr *) addr); //buf + 4 +4
		addrv4.sin_port = ntohs(*((uint16_t *) port)); //buf + 4 + 4+4
		// open socket (SOCK_NONBLOCK is not supported by old 2.6.26 kernel, therefore we do that manually)
		sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
		if (sock == -1)
			err(1, "socket() error, create UTP server socket");
        retval = fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK);
        if (retval != 0)
           err(1, "fcntl() error, set NONBLOCK flag on UTP server socket");
		// bind address
		retval = bind(sock, (const struct sockaddr *)&addrv4, sizeof(struct sockaddr_in));
		bind_errno = errno;
	} else {
		err(1, "binding utp sockets to IPv6 address not implemented yet.");
	}

	// notify simlib of newly created socket
	notify_real_fd(tmpFd, sock);

	char msg[TRIGGER_MSG_SIZE + 8];
	insert_type_fd(MUX_MSG_BIND_RET, sock, msg);
	insert_length(8, msg);
	(*(uint32_t *) (msg + TRIGGER_MSG_SIZE)) = retval;
	(*(uint32_t *) (msg + TRIGGER_MSG_SIZE + 4)) = bind_errno;
	control_write(msg, TRIGGER_MSG_SIZE + 8);

	// create new entry and append it to the list
	entry = create_new_listentry(sock, udp);
	entry->utp_ctx = ctx;
	fprintf(stderr, "created entry for UTP-UDP socket\n");
	set_read_handler(entry, handle_socket_read);
}
#endif

/*
 * establish a new client connection to this server
 */
static void mux_accept_internal(listentry * e) {
	if (e->prot == tcp) {

		int sock = accept(e->fd, NULL, NULL);

		if (sock > -1) {
			int retval = fcntl(sock, F_SETFL, fcntl(sock, F_GETFL) | O_NONBLOCK);
			if (retval != 0)
				err(1, "fcntl() error, set NONBLOCK flag");
			// append client connection
			listentry * entry = create_new_listentry(sock, tcp);
			set_handlers(entry, handle_socket_read, handle_socket_write);

			/* allocate for ipv6, should be sufficient for v4, too */
			unsigned int sockname_len = 3 * sizeof(int32_t) + sizeof(struct in6_addr);
			char sockname_buf[sockname_len];
			mux_getname_internal(sock, false, sockname_buf, &sockname_len);

			/* allocate for ipv6, should be sufficient for v4, too */
			unsigned int peername_len = 3 * sizeof(int32_t) + sizeof(struct in6_addr);
			char peername_buf[peername_len];
			mux_getname_internal(sock, true, peername_buf, &peername_len);

			// send reply message
			int msglen = TRIGGER_MSG_SIZE + 2 * sizeof(uint32_t) + sockname_len + peername_len;
			char msg[msglen];
			insert_type_fd(MUX_MSG_ACCEPT, e->fd, msg);
			insert_length(msglen - TRIGGER_MSG_SIZE, msg);
			*((size_t*) (msg + TRIGGER_MSG_SIZE)) = sock;
			*((size_t*) (msg + TRIGGER_MSG_SIZE + sizeof(uint32_t))) = errno;

			memcpy(msg + TRIGGER_MSG_SIZE + 2 * sizeof(uint32_t), sockname_buf, sockname_len);
			memcpy(msg + TRIGGER_MSG_SIZE + 2 * sizeof(uint32_t) + sockname_len, peername_buf, peername_len);

			control_write(msg, msglen);

		} else {
			// failed
			// send reply message
			int msglen = TRIGGER_MSG_SIZE + 2 * sizeof(uint32_t);
			char msg[msglen];
			insert_type_fd(MUX_MSG_ACCEPT, e->fd, msg);
			insert_length(2 * sizeof(uint32_t), msg);
			*((size_t*) (msg + TRIGGER_MSG_SIZE)) = sock;
			*((size_t*) (msg + TRIGGER_MSG_SIZE + sizeof(uint32_t))) = errno;
			control_write(msg, msglen);
		}
		set_read_handler(e, handle_new_listener);
	} else if(e->prot == utp){
		err(1, "This should not be possible to happen... seem to have received MSG_MUX_ACCEPT for UTP-socket, but non-auto accept cannot be supported");
	} else if(e->prot == udp){
		err(1, "This was not expected to happen... accepting a connection on a UDP socket");
	}
}

static void mux_accept(unsigned char * buf, int length) {
	assert_length(length, 4, "accept connection");
	// find matching entry
	listentry * e = entry_from_fd(*((uint32_t *) buf));
	if(e->prot == utp){
		err(1, "received MSG_MUX_ACCEPT for UTP-socket, but non-auto accept cannot be supported");
	}

	mux_accept_internal(e);
}

/*
 * closes an existing L4 connection
 *
 * buf is msg data as received from control dev (length 4)
 */
static void mux_close(unsigned char * buf, int length) {
	assert_length(length, FD_SIZE, "close_connection()");
#ifdef DO_UTP
	// for utp, use the UDP socket
	utp_listentry * uentry = uentry_from_usock((utp_socket*) *((uint32_t *) buf));
	if (uentry != 0) { //prot == utp
		errno = 0;
		utp_close(uentry->sock);
		//send close message
		send_trigger_msg(MUX_MSG_CLOSE, (uint32_t)(intptr_t)uentry->sock);

		// find previous entry and unlink e
		delete_utplistentry(uentry);
	} else
#endif
	{
		int retval;

		listentry * e = entry_from_fd(*((uint32_t *) buf));
		// close connection
		close: retval = close(e->fd);
		if (retval != 0) { // error
			if (errno == EINTR)
				goto close;
			// try again
			else
				err(1, "close failed");
		}
		//send close message
		send_trigger_msg(MUX_MSG_CLOSE, e->fd);

		// find previous entry and unlink e
		delete_listentry(e);
	}
}

/*
 * prepare the received data to write them later into the
 * previously opened and defined socket.
 */
static void mux_write(unsigned char * buf, int length) {
	size_t datalength;
	size_t retval;

	// length >= 8 is OK
	if (length < FD_SIZE)
		err(1, "muxwrite fails due to too short buf length()");

	datalength = length - FD_SIZE;
#ifdef DO_UTP
	// for utp, use the UDP socket
	utp_listentry * uentry = uentry_from_usock((utp_socket*) *((uint32_t *) buf));
	if (uentry != 0) { //prot == utp
		errno = 0;
		unsigned char*  current = buf;
		size_t sent_total = 0;
		while (current < buf+length) {
			size_t sent = utp_write(uentry->sock, current, datalength-sent_total);
			if (sent == 0) {
				break;
			}
			sent_total += sent;
			current += sent;
		}
		// send reply message
		char msg[TRIGGER_MSG_SIZE + 2 * sizeof(uint32_t)];
		insert_type_fd(MUX_MSG_WRITE_RET, (uintptr_t) uentry->sock, msg);
		insert_length(2 * sizeof(uint32_t), msg);
		*((size_t*) (msg + TRIGGER_MSG_SIZE)) = sent_total ;
		*((size_t*) (msg + TRIGGER_MSG_SIZE + sizeof(uint32_t))) = errno;
		control_write(msg, TRIGGER_MSG_SIZE + 2 * sizeof(uint32_t));
	} else
#endif
	{   // find matching entry
		listentry * e = entry_from_fd(*((uint32_t *) buf));

		// try to write
		retval = write(e->fd, buf + FD_SIZE, datalength);

		// send reply message
		char msg[TRIGGER_MSG_SIZE + 2 * sizeof(uint32_t)];
		insert_type_fd(MUX_MSG_WRITE_RET, e->fd, msg);
		insert_length(2 * sizeof(uint32_t), msg);
		*((size_t*) (msg + TRIGGER_MSG_SIZE)) = retval;
		*((size_t*) (msg + TRIGGER_MSG_SIZE + sizeof(uint32_t))) = errno;
		control_write(msg, TRIGGER_MSG_SIZE + 2 * sizeof(uint32_t));

		// signal again if it is still writable / becomes writable
		set_write_handler(e, handle_socket_write);
	}
}

static void mux_getsockopt(unsigned char * buf, int length) {
	//4 [fd] + 4[socket_level] + 4[option_name] + 4[buffer to create length (in byte)] = 16
	assert_length(length, 16, "invalid get_opts");

	int fd  = *((uint32_t *) buf);
	int32_t level = *((int32_t *) (buf + 4));
	int32_t optname = *((int32_t *) (buf + 8));
	socklen_t opt_length = *((uint32_t *) (buf + 12));

	int msg_size = TRIGGER_MSG_SIZE + opt_length + 2 * sizeof(int32_t);
	if (msg_size > MAX_ALLOCA_SIZE)
		fprintf(stderr, "WARNING: %d bytes for getsockopt() allocated on the stack\n", msg_size);
	char * msg = alloca(msg_size);
	errno = -1;
	int retval = getsockopt(fd, level, optname, (msg + TRIGGER_MSG_SIZE + 2 * sizeof(int32_t)), &opt_length);
	// opt_length is modified!!

	if (retval != 0)
		opt_length = 0;

	insert_type_fd(MUX_MSG_GETSOCKOPT, fd, msg);
	*((int32_t*)(msg + TRIGGER_MSG_SIZE)) = retval;
	*((int32_t*)(msg + TRIGGER_MSG_SIZE + sizeof(int32_t))) = errno;
	insert_length(opt_length + 2 * sizeof(int32_t), msg);
	control_write(msg, TRIGGER_MSG_SIZE + opt_length + 2*sizeof(int32_t));
}

static void mux_setsockopt(unsigned char * buf, int length) {
	if (length < 12)
		err(1, "invalid set_opts");

	int fd = *((uint32_t *) buf);
	int32_t level = *((int32_t *) (buf + 4));
	int32_t optname = *((int32_t *) (buf + 8));
	socklen_t opt_length = length - 12;

	int32_t retval = setsockopt(fd, level, optname, (buf + 12), opt_length);

	char msg[TRIGGER_MSG_SIZE + 2 * sizeof(int32_t)];
	insert_type_fd(MUX_MSG_SETSOCKOPT, fd, msg);
	insert_length(2 * sizeof(int32_t), msg);
	*((int32_t*) (msg + TRIGGER_MSG_SIZE + 0 * sizeof(int32_t))) = retval;
	*((int32_t*) (msg + TRIGGER_MSG_SIZE + 1 * sizeof(int32_t))) = errno;
	control_write(msg, TRIGGER_MSG_SIZE + 2 * sizeof(int32_t));
}

/* Writes getsockname() (peer == false) or getpeername() (peer == true) information into the given buffer.
 * If called for UTP and peer == true, fd has to be the utp_socket. If called for UTP and peer == false,
 * both utp_socket and UDP socket fd are supported. */
static void mux_getname_internal(int fd, bool peer, char buf[], unsigned int * length) {
	struct sockaddr_in6 addr;
	memset(&addr, 0, sizeof(addr));
	socklen_t addrlen = sizeof(addr);
	bool do_syscall = true;

#ifdef DO_UTP
	utp_listentry * uentry = uentry_from_usock((utp_socket *)fd);
	if (uentry != 0) {
		if(peer) {
			// use utp_getpeername instead of getpeername()
			utp_getpeername((utp_socket *)fd, (struct sockaddr*) &addr, &addrlen);
			do_syscall = false; // not required
		} else {
			// use normal getsockname(), but set fd to the UDP socket fd
			utp_context * ctx = uentry->utp_ctx;
			fd = udp_socket_from_utp_context(ctx);
		}
	}
#endif

	errno = -1;
	int retval = 0;
	if(do_syscall) {
		if(peer) {
			retval = getpeername(fd, (struct sockaddr*) &addr, &addrlen);
		} else {
			retval = getsockname(fd, (struct sockaddr*) &addr, &addrlen);
		}
	}

	if (addrlen > sizeof(addr)) {
		fprintf(stderr, "addrlen mismatch in getsockname, provided %i, would have got %i\n", (int)sizeof(struct sockaddr_in6), addrlen);
		fflush(stderr);
		exit(1);
	}

	// length depends on protocol
	unsigned int msglength;
	if (((struct sockaddr*) &addr)->sa_family == AF_INET) {
		msglength = 4 * sizeof(int32_t);
	} else if (((struct sockaddr*) &addr)->sa_family == AF_INET6) {
		msglength = 3 * sizeof(int32_t) + sizeof(struct in6_addr);
	} else {
		fprintf(stderr, "getsockname of sock %i returned neither AF_INET nor AF_INET6, but %i\n", fd, ((struct sockaddr*) &addr)->sa_family);
		fflush(stderr);
		exit(1);
	}

	// buffer should be allocated for IPv6, so it should always be sufficient
	if(*length < msglength) {
		fprintf(stderr, "buffer size too small: need %d, have %d\n", msglength, *length);
		fflush(stderr);
		exit(1);
	}

	// store length return value
	*length = msglength;

	// store output data into buffer
	*((int32_t*) (buf + 0 * sizeof(int32_t))) = retval;
	*((int32_t*) (buf + 1 * sizeof(int32_t))) = errno;
	if (((struct sockaddr*) &addr)->sa_family == AF_INET) {
		/* All IP addresses are exchanged in Big Endian == Java VM Byte Order  == Network Byte Order. No conversion required.
		 * Ports have to be exchanged in host byte order, conversion needed.
		 * */
		*((int32_t*) (buf + 2 * sizeof(int32_t))) = ((struct sockaddr_in*) &addr)->sin_addr.s_addr;
		*((int32_t*) (buf + 3 * sizeof(int32_t))) = htons(((struct sockaddr_in*) &addr)->sin_port);
	} else if (((struct sockaddr*) &addr)->sa_family == AF_INET6) {
		/* All IP addresses are exchanged in Big Endian == Java VM Byte Order  == Network Byte Order. No conversion required.
		 * Ports have to be exchanged in host byte order, conversion needed.
		 * */
		memcpy(buf + 2 * sizeof(int32_t), &addr.sin6_addr, sizeof(struct in6_addr));
		*((int32_t*) (buf + 2 * sizeof(int32_t) + sizeof(struct in6_addr))) = htons(addr.sin6_port);
	}
}

static void mux_getsockname(unsigned char * buf, int length) {
	assert_length(length, 4, "mux_getsockname()");
	int fd = *((uint32_t *) buf);

	/* allocate for ipv6, should be sufficient for v4, too */
	unsigned int msglength = 3 * sizeof(int32_t) + sizeof(struct in6_addr);
	char msg[TRIGGER_MSG_SIZE + msglength];

	mux_getname_internal(fd, false, msg + TRIGGER_MSG_SIZE, &msglength);

	insert_type_fd(MUX_MSG_GETSOCKNAME, fd, msg);
	insert_length(msglength, msg);

	control_write(msg, TRIGGER_MSG_SIZE + msglength);
}

static void mux_getpeername(unsigned char * buf, int length) {
	assert_length(length, 4, "mux_getpeername()");
	int fd = *((uint32_t *) buf);

	/* allocate for ipv6, should be sufficient for v4, too */
	unsigned int msglength = 3 * sizeof(int32_t) + sizeof(struct in6_addr);
	char msg[TRIGGER_MSG_SIZE + msglength];

	mux_getname_internal(fd, true, msg + TRIGGER_MSG_SIZE, &msglength);

	insert_type_fd(MUX_MSG_GETPEERNAME, fd, msg);
	insert_length(msglength, msg);

	control_write(msg, TRIGGER_MSG_SIZE + msglength);
}

static void mux_read_file(unsigned char * buf, int length) {
	/*
	 * buf contains 4 bytes dummy fd, 4 bytes maxlen and at least 2 bytes zero-terminated string
	 */
	if (length < 4 + 4 + 2)
		assert_length(length, 4 + 4 + 2, "mux_read_file()");

	int32_t maxlen = *((uint32_t *) (buf + 4));
	char * file = (char *) (buf + 8);
	char * msg = malloc(TRIGGER_MSG_SIZE + 4 * sizeof(int32_t) + maxlen);

	errno = -1;
	int fd = open(file, O_RDONLY);
	*((int32_t*) (msg + TRIGGER_MSG_SIZE + 0 * sizeof(int32_t))) = fd;
	*((int32_t*) (msg + TRIGGER_MSG_SIZE + 1 * sizeof(int32_t))) = errno;

	int bytes = 0;
	if (fd >= 0) {
		bool eof = false, error = false;
		while (bytes < maxlen && !eof && !error) {
			errno = -1;
			int retval = read(fd, msg + TRIGGER_MSG_SIZE + 4 * sizeof(int32_t) + bytes, maxlen - bytes);
			if (retval < 0) {
				error = true;
			} else if (retval == 0) {
				eof = true;
			} else {
				bytes += retval;
			}
		}

		*((int32_t*) (msg + TRIGGER_MSG_SIZE + 2 * sizeof(int32_t))) = error ? -1 : bytes;
		*((int32_t*) (msg + TRIGGER_MSG_SIZE + 3 * sizeof(int32_t))) = errno;

		int retval = close(fd);
		if (retval != 0)
			err(1, "error closing file %s\n", file);
	}
	insert_type_fd(MUX_MSG_READ_FILE, -1, msg);
	insert_length(4 * sizeof(int32_t) + bytes, msg);

	control_write(msg, TRIGGER_MSG_SIZE + 4 * sizeof(int32_t) + bytes);

	free(msg);
}

static void mux_write_file(unsigned char * buf, int length) {
	/*
	 * buf contains 4 bytes dummy fd, 4 bytes data offset, at least 2 bytes
	 * zero-terminated filename, and an undefined amount of data
	 */
	if (length < 4 + 4 + 2)
		assert_length(length, 4 + 4 + 2, "mux_write_file()");

	int32_t data_offset = *((uint32_t *) (buf + 4));
	char * file = (char *) (buf + 8);
	char * data = (char *) (buf + 8 + data_offset);
	int32_t data_len = length - 8 - data_offset;

	char msg[TRIGGER_MSG_SIZE + 4 * sizeof(int32_t)];
	insert_type_fd(MUX_MSG_WRITE_FILE, -1, msg);
	insert_length(4 * sizeof(int32_t), msg);

	errno = -1;
	int fd = open(file, O_WRONLY);

	*((int32_t*) (msg + TRIGGER_MSG_SIZE + 0 * sizeof(int32_t))) = fd;
	*((int32_t*) (msg + TRIGGER_MSG_SIZE + 1 * sizeof(int32_t))) = errno;

	int bytes = 0;
	if (fd >= 0) {
		bool error = false;
		while (bytes < data_len && !error) {
			errno = -1;
			int retval = write(fd, data + bytes, data_len - bytes);
			if (retval < 0) {
				error = true;
			} else {
				bytes += retval;
			}
		}
		*((int32_t*) (msg + TRIGGER_MSG_SIZE + 2 * sizeof(int32_t))) = error ? -1 : bytes;
		*((int32_t*) (msg + TRIGGER_MSG_SIZE + 3 * sizeof(int32_t))) = errno;

		int retval = close(fd);
		if (retval != 0)
			err(1, "error closing file %s\n", file);
	}

	control_write(msg, TRIGGER_MSG_SIZE + 4 * sizeof(int32_t));
}

static void main_loop(void) {
	struct epoll_event ev;
	int ret;
	while (true) {
#ifdef DO_UTP
		if (utp_ctx_head != NULL) {
			// this causes 500 ms timeout granularity for UTP retransmissions
			ret = epoll_wait(epollfd, &ev, 1, 500);
		} else
#endif
			// wait until something happens (timeout = infinity)
			ret = epoll_wait(epollfd, &ev, 1, -1);

		if (ret > 0) { //something happened
			listentry * e = ev.data.ptr;
			if (ev.events & (EPOLLIN | EPOLLRDHUP)) {
				if (e->readHandler == NULL) {
					fprintf(stderr, "epoll_wait() returned event EPOLLIN or EPOLLRDHUP, but no read handler is set\n");
				} else {
					e->readHandler(e);
				}
			}
			if (ev.events & EPOLLOUT) {
				if (e->writeHandler == NULL) {
					fprintf(stderr, "epoll_wait() returned event EPOLLOUT, but no write handler is set\n");
				} else {
					e->writeHandler(e);
				}
			}
			if(ev.events & EPOLLERR) {
				if (e->prot == control)
					err(1, "control connection broken, exiting");
				else {
					int error = 0;
					socklen_t errlen = sizeof(error);
					if (getsockopt(ev.data.fd, SOL_SOCKET, SO_ERROR, (void *)&error, &errlen) == 0)	{
					    printf("epoll_wait() returned error on non-control socket. Error was  %s\n", strerror(error));
					}
				}
			}
		} else if (ret < 0) { // an error occured
			if (errno != EINTR) {
				err(1, "epoll_wait() failed");
			}
		} // else nothing ready - try again (does this happen with infinite timeout?)
#ifdef DO_UTP
		if(utp_ctx_head != NULL) {
			utp_ctx_listentry* ctx_e;
			for (ctx_e = utp_ctx_head; ctx_e != NULL ; ctx_e = ctx_e->next) {
				utp_check_timeouts(ctx_e->utp_ctx);
			}
		}
#endif
	}
}

/*
 * read the input from the control device
 * @v buf data buffer where the read data are placed in
 * @v size amount of data which should be read
 * @ret 0 if the read succeded, MUX_INPUT_CLOSED if sender has closed the input or MUX_READ_ERROR if any error occured
 */
static int control_read_completely(void * buf, int minlen, int maxlen) {
	int bytes = 0;
	int retval;
	while (bytes < minlen) {
		retval = read(controlfd, buf + bytes, maxlen - bytes);
		if (retval >= 0) {
			bytes += retval;
		} else if ((errno == EAGAIN) || (errno == EINTR)) {
			// try again
		} else {
			fprintf(stderr, "FATAL: Error while reading (%s r=%d errno=%d )\n", strerror(errno), retval, errno);
			fflush(stderr);
			exit(-1);
		}
	}
	return bytes;
}

static void reset_rng();

static void handle_control(struct listentry * e) {
	static int messageCounter = 0;
	// receive single control message
//	unsigned char buf[MSGBUF_LENGTH] __attribute__ ((aligned (16)));
	unsigned char header_buf[sizeof(header)];

	messageCounter++;

	// read the type and the message length
	control_read_completely(header_buf, sizeof(header), sizeof(header));
	header.type = *((int32_t*) (header_buf + 0));
	header.length = *((int32_t*) (header_buf + 4));

    // calculate number of remainig bytes
	int total_size = sizeof(header) + header.length + DIGEST_LENGTH;
#if PADDING_BYTES > 1
    int size_with_padding;
    if ((total_size % PADDING_BYTES) != 0) {
        size_with_padding = ((total_size / PADDING_BYTES) + 1) * PADDING_BYTES;
    } else {
        size_with_padding = total_size;
    }
    //printf("adapted size of message %d from %d to %d bytes (%d bytes padding)\n", messageCounter, total_size, size_with_padding, size_with_padding - total_size);
    total_size = size_with_padding;
#endif
    total_size -= sizeof(header);

	// read remaining data
    unsigned char * data_buf = malloc(total_size);
	int read = control_read_completely(data_buf, total_size, total_size);

	if (read != total_size) {
		fprintf(stderr, "didn't read correct amount of data\n");
		fflush(stderr);
		exit(1);
	}

#ifdef DO_CRC
	if (!checkCRC(header_buf, sizeof(header), data_buf, header.length)) {
		printf("message %d: ", messageCounter);
		int i;
		for (i = 0; i < sizeof(header); i++)
		    printf("%hhd ", header_buf[i]);
		for (i = 0; i < header.length; i++)
		    printf("%hhd ", data_buf[i]);
		printf("\n received digest: ");
		for (i = 0; i < DIGEST_LENGTH; i++)
		    printf("%hhd ", data_buf[header.length + i]);
		printf("\n");
		fflush(stdout);
		fprintf(stderr, "invalid message digest\n");
		fflush(stderr);
		exit(1);
	}
#endif

	//execute the given control read part
	switch (header.type) {
	// open TCP connection
	case MUX_MSG_SOCKET_CONNECT:
		mux_tcp_socket_connect(data_buf, header.length);
		break;

		// close L4 connection
	case MUX_MSG_CLOSE:
		mux_close(data_buf, header.length);
		break;

		// write data to socket or buffer it
	case MUX_MSG_WRITE:
		mux_write(data_buf, header.length);
		break;

		// read data from socket and send back
	case MUX_MSG_READ:
		mux_read(data_buf, header.length);
		break;

	case MUX_MSG_SOCKET_BIND_LISTEN:
		mux_tcp_socket_bind_listen(data_buf, header.length);
		break;

	case MUX_MSG_ACCEPT:
		mux_accept(data_buf, header.length);
		break;

	case MUX_MSG_GETSOCKOPT:
		mux_getsockopt(data_buf, header.length);
		break;

	case MUX_MSG_SETSOCKOPT:
		mux_setsockopt(data_buf, header.length);
		break;

	case MUX_MSG_GETSOCKNAME:
		mux_getsockname(data_buf, header.length);
		break;

	case MUX_MSG_GETPEERNAME:
		mux_getpeername(data_buf, header.length);
		break;

	case MUX_MSG_READ_FILE:
		mux_read_file(data_buf, header.length);
		break;

	case MUX_MSG_WRITE_FILE:
		mux_write_file(data_buf, header.length);
		break;

	case MUX_MSG_RESET_RNG:
		reset_rng();
		break;
#ifdef DO_UTP
	// open uTP connection
	case MUX_MSG_UTP_SOCKET_CONNECT:
		mux_utp_socket_connect(data_buf, header.length);
		break;
	case MUX_MSG_UTP_SOCKET_BIND_LISTEN:
		mux_utp_socket_bind_listen(data_buf, header.length);
		break;
#endif
	default:
		fprintf(stderr, "unknown message type %d\n", header.type);
		fflush(stderr);
		exit(-1);
	}

	free(data_buf);
}

static void write_startup_msg() {
	printf("sending startup message\n");
	fflush(stdout);
	send_trigger_msg(MUX_MSG_START, -1);
}

static void reset_single_pool(off_t addr, int data_size) {
	// from linux/drivers/char/random.c
#define EXTRACT_SIZE 10
	struct entropy_store {
		/* read-only data: */
		struct poolinfo *poolinfo;
		__u32 *pool;
		const char *name;
		struct entropy_store *pull;
		int limit;

		/* read-write data: */
		int padding___lock;
		unsigned add_ptr;
		unsigned input_rotate;
		int entropy_count;
		int entropy_total;
		unsigned int initialized :1;
		__u8 last_data[EXTRACT_SIZE];
	};

	int pagesize = sysconf(_SC_PAGE_SIZE);
	off_t page = (addr >> 12) << 12;
	off_t offset = addr - page;

	int fd = open("/dev/kmem", O_RDWR);
	if (fd < 0)
		err(1, "opening of /dev/kmem failed");

	void * buf = mmap(NULL, pagesize, PROT_READ | PROT_WRITE, MAP_SHARED, fd, page);
	if (buf == MAP_FAILED)
		err(1, "mmap failed\n");

	struct entropy_store * input_pool = (struct entropy_store *) (buf + offset);
	input_pool->add_ptr = 0;
	input_pool->input_rotate = 0;
	input_pool->entropy_count = 0;
	input_pool->entropy_total = 0;
	input_pool->initialized = 1;

	off_t dat_page = (((unsigned int)(intptr_t) input_pool->pool) >> 12) << 12;
	off_t dat_offset = ((unsigned int)(intptr_t) input_pool->pool) - dat_page;

	munmap(buf, pagesize);

	buf = mmap(NULL, pagesize, PROT_READ | PROT_WRITE, MAP_SHARED, fd, dat_page);
	if (buf == MAP_FAILED)
		err(1, "mmap(dat) failed\n");

	uint8_t * pool_data = buf + dat_offset;
	int i;
	for (i = 0; i < data_size; i++)
		pool_data[i] = 0;

	munmap(buf, pagesize);

	close(fd);
}

static void reset_fast_pool(off_t addr) {
	// from linux/drivers/char/random.c
	struct fast_pool {
		__u32 pool[4];
		unsigned long last;
		unsigned short count;
		unsigned char rotate;
		unsigned char last_timer_intr;
	};
	off_t page = (addr >> 12) << 12;
	off_t offset = addr - page;
	int pagesize = sysconf(_SC_PAGE_SIZE);

	int fd = open("/dev/kmem", O_RDWR);
	if (fd < 0)
		err(1, "opening of /dev/kmem failed");

	void * buf = mmap(NULL, pagesize, PROT_READ | PROT_WRITE, MAP_SHARED, fd, page);
	if (buf == MAP_FAILED)
		err(1, "mmap failed\n");

	struct fast_pool * p = (struct fast_pool *) (buf + offset);
	p->pool[0] = 0;
	p->pool[1] = 0;
	p->pool[2] = 0;
	p->pool[3] = 0;
	p->last = 0;
	p->count = 0;
	p->rotate = 0;
	p->last_timer_intr = 0;

	munmap(buf, pagesize);
	close(fd);
}

/*
 * Resets the system RNG in the kernel so that a VM does always have the same
 * random numbers.
 *
 * FIXME this method does not work. There seems to be 1 bit of randomness left,
 * resulting in two alternate streams of random numbers. In addition, the random
 * data of each host seems to be different, even if this method is called from a
 * SimLib event at the same time. Could that be the HPETs fault?
 *
 * see linux/drivers/char/random.c
 */
static void reset_rng() {
	// from linux/drivers/char/random.c
#define INPUT_POOL_WORDS 128
#define OUTPUT_POOL_WORDS 32
	// addresses from System.map
	reset_single_pool(0xc124ada0, INPUT_POOL_WORDS * 4); //input_pool
	reset_single_pool(0xc124ade0, OUTPUT_POOL_WORDS * 4); //blocking_pool
	reset_single_pool(0xc124ae20, OUTPUT_POOL_WORDS * 4); //nonblocking_pool
	reset_fast_pool(0xc124ae64); //irq_randomness
}

int main(int argc, char **argv) {
	if (argc != 2) {
		fprintf(stderr, "usage: %s <device>\n", argv[0]);
		fflush(stderr);
		exit(1);
	}
	char* device = argv[1];

	controlfd = open(device, O_RDWR | O_CLOEXEC | O_APPEND | O_NOCTTY);
	if (controlfd == -1) {
		err(1, "unable to open control device %s", device);
	}

	//change to raw input and output
	struct termios options;
	tcgetattr(controlfd, &options);
	options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG); // raw mode on serial device
	options.c_iflag &= ~(ICRNL | INLCR | IGNCR); // do not map between <CR> and <NL> and do not ignore the <cr>
	options.c_oflag &= ~OPOST;
	tcsetattr(controlfd, TCSANOW, &options);

	struct serial_struct serial;
	ioctl(controlfd, TIOCGSERIAL, &serial);
	serial.flags |= ASYNC_LOW_LATENCY; // (0x2000)
	ioctl(controlfd, TIOCSSERIAL, &serial);

	// create epoll fd
	// parameter not used for current kernels, but epoll_create1() not supported by 2.6.26 kernel
	epollfd = epoll_create(10000);
	if(epollfd < 0)
		err(1, "creating epoll fd failed");
	// create list entry for control fd
	tcp_head = NULL;
	listentry * e = create_new_listentry(controlfd, control);
	set_read_handler(e, handle_control);
	e->autoRead = false;
	e->autoAccept = false;

	write_startup_msg();
	main_loop();
	exit(0); // should not happen
}
